#
# coding=utf-8
"""nmea - Imports GPS NMEA-formatted data files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import logging

from functools import reduce
from operator import xor

from upoints import (point, utils)


def calc_checksum(sentence):
    """Calculate a NMEA 0183 checksum for the given sentence.

    NMEA checksums are a simple XOR of all the characters in the sentence
    between the leading "$" symbol, and the "*" checksum separator.

    :param str sentence: NMEA 0183 formatted sentence

    """
    if sentence.startswith('$'):
        sentence = sentence[1:]
    sentence = sentence.split('*')[0]
    return reduce(xor, map(ord, sentence))


def nmea_latitude(latitude):
    """Generate a NMEA-formatted latitude pair.

    :param float latitude: Latitude to convert
    :rtype: ``tuple``
    :return: NMEA-formatted latitude values

    """
    return ('%02i%07.4f' % utils.to_dms(abs(latitude), 'dm'),
            'N' if latitude >= 0 else 'S')


def nmea_longitude(longitude):
    """Generate a NMEA-formatted longitude pair.

    :param float longitude: Longitude to convert
    :rtype: ``tuple``
    :return: NMEA-formatted longitude values

    """
    return ('%03i%07.4f' % utils.to_dms(abs(longitude), 'dm'),
            'E' if longitude >= 0 else 'W')


def parse_latitude(latitude, hemisphere):
    """Parse a NMEA-formatted latitude pair.

    :param str latitude: Latitude in DDMM.MMMM
    :param str hemisphere: North or South
    :rtype: ``float``
    :return: Decimal representation of latitude

    """
    latitude = int(latitude[:2]) + float(latitude[2:]) / 60
    if hemisphere == 'S':
        latitude = -latitude
    elif not hemisphere == 'N':
        raise ValueError('Incorrect North/South value %r' % hemisphere)
    return latitude


def parse_longitude(longitude, hemisphere):
    """Parse a NMEA-formatted longitude pair.

    :param str longitude: Longitude in DDDMM.MMMM
    :param str hemisphere: East or West
    :rtype: ``float``
    :return: Decimal representation of longitude

    """
    longitude = int(longitude[:3]) + float(longitude[3:]) / 60
    if hemisphere == 'W':
        longitude = -longitude
    elif not hemisphere == 'E':
        raise ValueError('Incorrect North/South value %r' % hemisphere)
    return longitude

#: NMEA's mapping of code to reading type
MODE_INDICATOR = {
    'A': 'Autonomous',
    'D': 'Differential',
    'E': 'Estimated',
    'M': 'Manual',
    'S': 'Simulated',
    'N': 'Invalid',
}


class LoranPosition(point.Point):

    """Class for representing a GPS NMEA-formatted Loran-C position."""

    __slots__ = ('time', 'status', 'mode')

    def __init__(self, latitude, longitude, time, status, mode=None):
        """Initialise a new ``LoranPosition`` object.

        :param float latitude: Fix's latitude
        :param float longitude: Fix's longitude
        :param datetime.time time: Time the fix was taken
        :param bool status: Whether the data is active
        :param str mode: Type of reading

        """
        super(LoranPosition, self).__init__(latitude, longitude)
        self.time = time
        self.status = status
        self.mode = mode

    def __str__(self, talker='GP'):
        """Pretty printed position string.

        :param str talker: Talker ID
        :rtype: ``str``
        :return: Human readable string representation of ``Position`` object

        """
        if not len(talker) == 2:
            raise ValueError('Talker ID must be two characters %r' % talker)
        data = ['%sGLL' % talker]
        data.extend(nmea_latitude(self.latitude))
        data.extend(nmea_longitude(self.longitude))
        data.append('%s.%02i' % (self.time.strftime('%H%M%S'),
                                 self.time.microsecond / 1000000))
        data.append('A' if self.status else 'V')
        if self.mode:
            data.append(self.mode)
        data = ','.join(data)
        return '$%s*%02X\r' % (data, calc_checksum(data))

    def mode_string(self):
        """Return a string version of the reading mode information.

        :rtype: ``str``
        :return: Quality information as string

        """
        return MODE_INDICATOR.get(self.mode, 'Unknown')

    @staticmethod
    def parse_elements(elements):
        """Parse position data elements.

        :param list elements: Data values for fix
        :rtype: ``Fix``
        :return: Fix object representing data

        """
        if not len(elements) in (6, 7):
            raise ValueError('Invalid GLL position data')
        # Latitude and longitude are checked for validity during Fix
        # instantiation
        latitude = parse_latitude(elements[0], elements[1])
        longitude = parse_longitude(elements[2], elements[3])
        hour, minute, second = [int(elements[4][i:i + 2])
                                for i in range(0, 6, 2)]
        usecond = int(elements[4][6:8]) * 10000
        time = datetime.time(hour, minute, second, usecond)
        active = True if elements[5] == 'A' else False
        mode = elements[6] if len(elements) == 7 else None
        return LoranPosition(latitude, longitude, time, active, mode)


class Position(point.Point):

    """Class for representing a GPS NMEA-formatted position.

    .. versionadded:: 0.8.0

    """

    __slots__ = ('time', 'status', 'speed', 'track', 'date', 'variation',
                 'mode')

    def __init__(self, time, status, latitude, longitude, speed, track, date,
                 variation, mode=None):
        """Initialise a new ``Position`` object.

        :param datetime.time time: Time the fix was taken
        :param bool status: Whether the data is active
        :param float latitude: Fix's latitude
        :param float longitude: Fix's longitude
        :param float speed: Ground speed
        :param float track: Track angle
        :param datetime.date date: Date when position was taken
        :param float variation: Magnetic variation
        :param str mode: Type of reading

        """
        super(Position, self).__init__(latitude, longitude)
        self.time = time
        self.status = status
        self.speed = speed
        self.track = track
        self.date = date
        self.variation = variation
        self.mode = mode

    def __str__(self):
        """Pretty printed position string.

        :rtype: ``str``
        :return: Human readable string representation of ``Position`` object

        """
        data = ['GPRMC']
        data.append(self.time.strftime('%H%M%S'))
        data.append('A' if self.status else 'V')
        data.extend(nmea_latitude(self.latitude))
        data.extend(nmea_longitude(self.longitude))
        data.append('%.1f' % self.speed)
        data.append('%.1f' % self.track)
        data.append(self.date.strftime('%d%m%y'))
        if self.variation == int(self.variation):
            data.append('%i' % abs(self.variation))
        else:
            data.append('%.1f' % abs(self.variation))
        data.append('E' if self.variation >= 0 else 'W')
        if self.mode:
            data.append(self.mode)
        data = ','.join(data)
        return '$%s*%02X\r' % (data, calc_checksum(data))

    def mode_string(self):
        """Return a string version of the reading mode information.

        :rtype: ``str``
        :return: Quality information as string

        """
        return MODE_INDICATOR.get(self.mode, 'Unknown')

    @staticmethod
    def parse_elements(elements):
        """Parse position data elements.

        :param list elements: Data values for position
        :rtype: ``Position``
        :return: Position object representing data

        """
        if not len(elements) in (11, 12):
            raise ValueError('Invalid RMC position data')
        time = datetime.time(*[int(elements[0][i:i + 2])
                               for i in range(0, 6, 2)])
        active = True if elements[1] == 'A' else False
        # Latitude and longitude are checked for validity during Fix
        # instantiation
        latitude = parse_latitude(elements[2], elements[3])
        longitude = parse_longitude(elements[4], elements[5])
        speed = float(elements[6])
        track = float(elements[7])
        date = datetime.date(2000 + int(elements[8][4:6]),
                             int(elements[8][2:4]), int(elements[8][:2]))
        variation = float(elements[9]) if not elements[9] == '' else None
        if elements[10] == 'W':
            variation = -variation
        elif variation and not elements[10] == 'E':
            raise ValueError('Incorrect variation value %r'
                             % elements[10])
        mode = elements[11] if len(elements) == 12 else None
        return Position(time, active, latitude, longitude, speed, track, date,
                        variation, mode)


class Fix(point.Point):

    """Class for representing a GPS NMEA-formatted system fix.

    .. versionadded:: 0.8.0

    """

    __slots__ = ('time', 'quality', 'satellites', 'dilution', 'altitude',
                 'geoid_delta', 'dgps_delta', 'dgps_station', 'mode')

    fix_quality = [
        'Invalid',
        'GPS',
        'DGPS',
        'PPS',
        'Real Time Kinematic'
        'Float RTK',
        'Estimated',
        'Manual',
        'Simulation',
    ]

    def __init__(self, time, latitude, longitude, quality, satellites,
                 dilution, altitude, geoid_delta, dgps_delta=None,
                 dgps_station=None, mode=None):
        """Initialise a new ``Fix`` object.

        :param datetime.time time: Time the fix was taken
        :param float latitude: Fix's latitude
        :param float longitude: Fix's longitude
        :param int quality: Mode under which the fix was taken
        :param int satellites: Number of tracked satellites
        :param float dilution: Horizontal dilution at reported position
        :param float altitude: Altitude above MSL
        :param float geoid_delta: Height of geoid's MSL above WGS84 ellipsoid
        :param float dgps_delta: Number of seconds since last DGPS sync
        :param int dgps_station: Identifier of the last synced DGPS station
        :param str mode: Type of reading

        """
        super(Fix, self).__init__(latitude, longitude)
        self.time = time
        self.quality = quality
        self.satellites = satellites
        self.dilution = dilution
        self.altitude = altitude
        self.geoid_delta = geoid_delta
        self.dgps_delta = dgps_delta
        self.dgps_station = dgps_station
        self.mode = mode

    def __str__(self):
        """Pretty printed location string.

        :rtype: ``str``
        :return: Human readable string representation of ``Fix`` object

        """
        data = ['GPGGA']
        data.append(self.time.strftime('%H%M%S'))
        data.extend(nmea_latitude(self.latitude))
        data.extend(nmea_longitude(self.longitude))
        data.append(str(self.quality))
        data.append('%02i' % self.satellites)
        data.append('%.1f' % self.dilution)
        data.append('%.1f' % self.altitude)
        data.append('M')
        data.append('-' if not self.geoid_delta else '%.1f' % self.geoid_delta)
        data.append('M')
        data.append('%.1f' % self.dgps_delta if self.dgps_delta else '')
        data.append('%04i' % self.dgps_station if self.dgps_station else '')
        data = ','.join(data)
        return '$%s*%02X\r' % (data, calc_checksum(data))

    def quality_string(self):
        """Return a string version of the quality information.

        :rtype: ``str``
        :return: Quality information as string

        """
        return self.fix_quality[self.quality]

    @staticmethod
    def parse_elements(elements):
        """Parse essential fix's data elements.

        :param list elements: Data values for fix
        :rtype: ``Fix``
        :return: Fix object representing data

        """
        if not len(elements) in (14, 15):
            raise ValueError('Invalid GGA fix data')
        time = datetime.time(*[int(elements[0][i:i + 2])
                               for i in range(0, 6, 2)])
        # Latitude and longitude are checked for validity during Fix
        # instantiation
        latitude = parse_latitude(elements[1], elements[2])
        longitude = parse_longitude(elements[3], elements[4])
        quality = int(elements[5])
        if not 0 <= quality <= 9:
            raise ValueError('Invalid quality value %r' % quality)
        satellites = int(elements[6])
        if not 0 <= satellites <= 12:
            raise ValueError('Invalid number of satellites %r'
                             % satellites)
        dilution = float(elements[7])
        altitude = float(elements[8])
        if elements[9] == 'F':
            altitude = altitude * 3.2808399
        elif not elements[9] == 'M':
            raise ValueError('Unknown altitude unit %r' % elements[9])
        if elements[10] in ('-', ''):
            geoid_delta = False
            logging.warning('Altitude data could be incorrect, as the geoid '
                            'difference has not been provided')
        else:
            geoid_delta = float(elements[10])
        if elements[11] == 'F':
            geoid_delta = geoid_delta * 3.2808399
        elif geoid_delta and not elements[11] == 'M':
            raise ValueError('Unknown geoid delta unit %r' % elements[11])
        dgps_delta = float(elements[12]) if elements[12] else None
        dgps_station = int(elements[13]) if elements[13] else None
        mode = elements[14] if len(elements) == 15 else None
        return Fix(time, latitude, longitude, quality, satellites, dilution,
                   altitude, geoid_delta, dgps_delta, dgps_station, mode)


class Waypoint(point.Point):

    """Class for representing a NMEA-formatted waypoint.

    .. versionadded:: 0.8.0

    """

    __slots__ = ('name', )

    def __init__(self, latitude, longitude, name):
        """Initialise a new ``Waypoint`` object.

        :param float latitude: Waypoint's latitude
        :param float longitude: Waypoint's longitude
        :param str name: Comment for waypoint

        """
        super(Waypoint, self).__init__(latitude, longitude)
        self.name = name.upper()

    def __str__(self):
        """Pretty printed location string.

        :rtype: ``str``
        :return: Human readable string representation of ``Waypoint`` object

        """
        data = ['GPWPL']
        data.extend(nmea_latitude(self.latitude))
        data.extend(nmea_longitude(self.longitude))
        data.append(self.name)
        data = ','.join(data)
        text = '$%s*%02X\r' % (data, calc_checksum(data))
        if len(text) > 81:
            raise ValueError('All NMEA sentences must be less than 82 bytes '
                             'including line endings')
        return text

    @staticmethod
    def parse_elements(elements):
        """Parse waypoint data elements.

        :param list elements: Data values for fix
        :rtype: ``Waypoint``
        :return: ``Waypoint`` object representing data

        """
        if not len(elements) == 5:
            raise ValueError('Invalid WPL waypoint data')
        # Latitude and longitude are checked for validity during Fix
        # instantiation
        latitude = parse_latitude(elements[0], elements[1])
        longitude = parse_longitude(elements[2], elements[3])
        name = elements[4]
        return Waypoint(latitude, longitude, name)


class Locations(point.Points):

    """Class for representing a group of GPS location objects.

    .. versionadded:: 0.8.0

    """

    def __init__(self, gpsdata_file=None):
        """Initialise a new ``Locations`` object."""
        super(Locations, self).__init__()
        self._gpsdata_file = gpsdata_file
        if gpsdata_file:
            self.import_locations(gpsdata_file)

    def import_locations(self, gpsdata_file, checksum=True):
        r"""Import GPS NMEA-formatted data files

        ``import_locations()`` returns a list of `Fix` objects representing the
        fix sentences found in the GPS data.

        It expects data files in NMEA 0183 format, as specified in `the
        official documentation`_, which is ASCII text such as::

            $GPGSV,6,6,21,32,65,170,35*48
            $GPGGA,142058,5308.6414,N,00300.9257,W,1,04,5.6,1374.6,M,34.5,M,,*6B
            $GPRMC,142058,A,5308.6414,N,00300.9257,W,109394.7,202.9,191107,5,E,A*2C
            $GPGSV,6,1,21,02,76,044,43,03,84,156,49,06,89,116,51,08,60,184,30*7C
            $GPGSV,6,2,21,09,87,321,50,10,77,243,44,11,85,016,49,12,89,100,52*7A
            $GPGSV,6,3,21,13,70,319,39,14,90,094,52,16,85,130,49,17,88,136,51*7E
            $GPGSV,6,4,21,18,57,052,27,24,65,007,34,25,62,142,32,26,88,031,51*73
            $GPGSV,6,5,21,27,64,343,33,28,45,231,16,30,84,198,49,31,90,015,52*7C
            $GPGSV,6,6,21,32,65,170,34*49
            $GPWPL,5200.9000,N,00013.2600,W,HOME*5E
            $GPGGA,142100,5200.9000,N,00316.6600,W,1,04,5.6,1000.0,M,34.5,M,,*68
            $GPRMC,142100,A,5200.9000,N,00316.6600,W,123142.7,188.1,191107,5,E,A*21

        The reader only imports the GGA, or GPS fix, sentences currently but
        future versions will probably support tracks and waypoints.  Other than
        that the data is out of scope for ``upoints``.

        The above file when processed by ``import_locations()`` will return the
        following ``list`` object::

            [Fix(datetime.time(14, 20, 58), 53.1440233333, -3.01542833333, 1,
                 4, 5.6, 1374.6, 34.5, None, None),
             Position(datetime.time(14, 20, 58), True, 53.1440233333,
                      -3.01542833333, 109394.7, 202.9,
                      datetime.date(2007, 11, 19), 5.0, 'A'),
             Waypoint(52.015, -0.221, 'Home'),
             Fix(datetime.time(14, 21), 52.015, -3.27766666667, 1, 4, 5.6,
                 1000.0, 34.5, None, None),
             Position(datetime.time(14, 21), True, 52.015, -3.27766666667,
                      123142.7, 188.1, datetime.date(2007, 11, 19), 5.0, 'A')]

        .. note::
           The standard is quite specific in that sentences *must* be less than
           82 bytes, while it would be nice to add yet another validity check
           it isn't all that uncommon for devices to break this requirement in
           their "extensions" to the standard.
        .. todo:: Add optional check for message length, on by default

        :type gpsdata_file: ``file``, ``list`` or ``str``
        :param gpsdata_file: NMEA data to read
        :param bool checksum: Whether checksums should be tested
        :rtype: ``list``
        :return: Series of locations taken from the data

        .. _the official documentation: http://en.wikipedia.org/wiki/NMEA_0183

        """
        self._gpsdata_file = gpsdata_file
        data = utils.prepare_read(gpsdata_file)

        parsers = {
            'GPGGA': Fix,
            'GPRMC': Position,
            'GPWPL': Waypoint,
            'GPGLL': LoranPosition,
            'LCGLL': LoranPosition,
        }

        if not checksum:
            logging.warning('Disabling the checksum tests should only be used'
                            'when the device is incapable of emitting the '
                            'correct values!')
        for line in data:
            # The standard tells us lines should end in \r\n even though some
            # devices break this, but Python's standard file object solves this
            # for us anyway.  However, be careful if you implement your own
            # opener.
            if not line[1:6] in parsers:
                continue
            if checksum:
                values, checksum = line[1:].split('*')
                if not calc_checksum(values) == int(checksum, 16):
                    raise ValueError('Sentence has invalid checksum')
            else:
                values = line[1:].split('*')[0]
            elements = values.split(',')
            parser = getattr(parsers[elements[0]], 'parse_elements')
            self.append(parser(elements[1:]))



#
# coding=utf-8
"""test_kml - Test kml support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import TestCase

from expecter import expect

from upoints.kml import (Placemark, Placemarks, etree)

from tests.utils import xml_compare


class TestPlacemark(TestCase):
    def test___repr__(self):
        expect(Placemark(52, 0, 4)) == \
            'Placemark(52.0, 0.0, 4.0, None, None)'
        expect(Placemark(52, 0, None)) == \
            'Placemark(52.0, 0.0, None, None, None)'
        expect(Placemark(52, 0, None, 'name', 'desc')) == \
            "Placemark(52.0, 0.0, None, 'name', 'desc')"

    def test___str__(self):
        expect(str(Placemark(52, 0, 4))) == \
            """52°00'00"N, 000°00'00"E alt 4m"""
        expect(str(Placemark(52, 0, None))) == \
            """52°00'00"N, 000°00'00"E"""
        expect(str(Placemark(52, 0, None, 'name', 'desc'))) == \
            """name (52°00'00"N, 000°00'00"E) [desc]"""
        expect(str(Placemark(52, 0, 42, 'name', 'desc'))) == \
            """name (52°00'00"N, 000°00'00"E alt 42m) [desc]"""

    def test_tokml(self):
        expect(etree.tostring(Placemark(52, 0, 4).tokml())) == \
            b'<kml:Placemark xmlns:kml="http://earth.google.com/kml/2.2">' \
            b'<kml:Point><kml:coordinates>0.0,52.0,4</kml:coordinates></kml:Point>' \
            b'</kml:Placemark>'
        expect(etree.tostring(Placemark(52, 0, 4, 'Cambridge').tokml())) == \
            b'<kml:Placemark xmlns:kml="http://earth.google.com/kml/2.2" id="Cambridge">' \
            b'<kml:name>Cambridge</kml:name><kml:Point><kml:coordinates>0.0,52.0,4</kml:coordinates></kml:Point>' \
            b'</kml:Placemark>'
        expect(etree.tostring(Placemark(52, 0, 4).tokml())) == \
            b'<kml:Placemark xmlns:kml="http://earth.google.com/kml/2.2">' \
            b'<kml:Point><kml:coordinates>0.0,52.0,4</kml:coordinates></kml:Point></kml:Placemark>'
        expect(etree.tostring(Placemark(52, 0, 4, 'Cambridge', 'in the UK').tokml())) == \
            b'<kml:Placemark xmlns:kml="http://earth.google.com/kml/2.2" id="Cambridge">' \
            b'<kml:name>Cambridge</kml:name><kml:description>in the UK</kml:description>' \
            b'<kml:Point><kml:coordinates>0.0,52.0,4</kml:coordinates></kml:Point>' \
            b'</kml:Placemark>'


class TestPlacemarks(TestCase):
    def test_import_locations(self):
        locations = Placemarks(open('tests/data/kml'))
        expect(str(locations['Cambridge'])) ==  \
            """Cambridge (52°10'01"N, 000°23'24"E)"""
        expect(str(locations['Home'])) == \
            """Home (52°00'54"N, 000°13'15"W alt 60m)"""

    def test_export_kml_file(self):
        locations = Placemarks(open('tests/data/kml'))
        export = locations.export_kml_file()
        kml_xml = etree.parse('tests/data/kml')
        for e1, e2 in zip(export.getiterator(), kml_xml.getiterator()):
            xml_compare(e1, e2)



#
# coding=utf-8
"""test_tzdata - Test tzdata support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from operator import attrgetter
from unittest import TestCase

from expecter import expect

from upoints.tzdata import (Zone, Zones)


class TestZone(TestCase):
    def test___repr__(self):
        expect(repr(Zone('+513030-0000731', 'GB', 'Europe/London'))) == \
            "Zone('+513030-0000730', 'GB', 'Europe/London', None)"

    def test___str__(self):
        expect(str(Zone('+513030-0000731', 'GB', 'Europe/London'))) == \
            """Europe/London (GB: 51°30'30"N, 000°07'30"W)"""
        expect(str(Zone('+0658-15813', 'FM', 'Pacific/Ponape',
                        ['Ponape (Pohnpei)', ]))) == \
            """Pacific/Ponape (FM: 06°58'00"N, 158°13'00"W also Ponape (Pohnpei))"""


class TestZones(TestCase):
    def test_import_locations(self):
        zones = Zones(open('tests/data/timezones'))
        data = [str(v) for v in sorted(zones, key=attrgetter('zone'))]
        expect(data[0]) == """Africa/Luanda (AO: 08°48'00"S, 013°14'00"E)"""
        expect(data[1]) == """America/Curacao (AN: 12°11'00"N, 069°00'00"W)"""
        expect(data[2]) == \
            """Antarctica/McMurdo (AQ: 77°50'00"S, 166°36'00"E also McMurdo Station, Ross Island)"""

    def test_dump_zone_file(self):
        zones = Zones(open('tests/data/timezones'))
        expect(Zones.dump_zone_file(zones)) == \
            ['AN\t+121100-0690000\tAmerica/Curacao',
             'AO\t-084800+0131400\tAfrica/Luanda',
             'AQ\t-775000+1663600\tAntarctica/McMurdo\tMcMurdo Station, Ross Island']



#
# coding=utf-8
"""test_baken - Test baken support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import TestCase

from expecter import expect

from upoints.baken import (Baken, Bakens)


class TestBaken(TestCase):
    def test___repr__(self):
        expect(repr(Baken(14.460, 20.680, None, None, None, 0.000, None, None,
                          None, None, None))) == \
            'Baken(14.46, 20.68, None, None, None, 0.0, None, None, None, None, None)'
        expect(repr(Baken(None, None, '2 x Turnstile', None, 50.000, 460.000,
                          'IO93BF', 'A1A', None, 25, None))) == \
            ("Baken(%s, -1.875, '2 x Turnstile', None, 50.0, "
             "460.0, 'IO93BF', 'A1A', None, 25, None)" % 53.229166666666686)

        with expect.raises(LookupError,
                           ('Unable to instantiate baken object, no latitude '
                            'or locator string')):
            Baken(None, None)

    def test__set_locator(self):
        test = Baken(None, None, '2 x Turnstile', None, 50.000, 460.000,
                     'IO93BF', 'A1A', None, 25, None)
        test.locator = 'JN44FH'
        expect(test.latitude) == 44.3125
        expect(test.longitude) == 8.458333333333314

    def test___str__(self):
        expect(str(Baken(14.460, 20.680, None, None, None, 0.000, None, None,
                         None, None, None))) == \
            """14°27'36"N, 020°40'48"E"""
        expect(str(Baken(None, None, '2 x Turnstile', None, 50.000, 460.000,
                         'IO93BF', 'A1A', None, 25, None))) == \
            """IO93BF (53°13'45"N, 001°52'30"W)"""


class TestBakens(TestCase):
    def test_import_locations(self):
        locations = Bakens(open('tests/data/baken_data'))
        data = ['%s - %s' % (k, v) for k, v in sorted(locations.items())]
        expect(data[0]) == """Abeche, Chad - 14°27'36"N, 020°40'48"E"""
        expect(data[1]) == """GB3BUX - IO93BF (53°13'45"N, 001°52'30"W)"""
        expect(data[2]) == """IW1RCT - JN44FH (44°18'45"N, 008°27'29"E)"""
        locations = Bakens(open('tests/data/no_valid_baken'))
        expect(len(locations)) == 0



#
# coding=utf-8
"""test_cellid - Test cellid support"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from unittest import TestCase

from expecter import expect

import datetime

from upoints.cellid import (Cell, Cells)


class TestCell(TestCase):
    def test___repr__(self):
        expect(repr(Cell(4, 52.015, -0.221, 21, 46, 40000, 10, 0, 1,
                         datetime.datetime(2008, 4, 15, 15, 21, 35),
                         datetime.datetime(2008, 4, 15, 15, 28, 49)))) == \
            ('Cell(4, 52.015, -0.221, 21, 46, 40000, 10, 0, 1, '
             'datetime.datetime(2008, 4, 15, 15, 21, 35), '
             'datetime.datetime(2008, 4, 15, 15, 28, 49))')

    def test___str__(self):
        expect(str(Cell(4, 52.015, -0.221, 21, 46, 40000, 10, 0, 1,
                        datetime.datetime(2008, 4, 15, 15, 21, 35),
                        datetime.datetime(2008, 4, 15, 15, 28, 49)))) == \
            ('4,52.0150000000000,-0.2210000000000,21,46,40000,10,0,1,'
             '2008-04-15 15:21:35,2008-04-15 15:28:49')


class TestCells(TestCase):
    def setUp(self):
        self.cells = Cells(open('tests/data/cells'))

    def test___str__(self):
        data = sorted(map(str, self.cells.values()))
        expect(data[0]) == \
            ('22747,52.0438995361328,-0.2246370017529,234,33,2319,647,0,1,'
             '2008-04-05 21:32:40,2008-04-05 21:32:40')
        expect(data[1]) == \
            ('22995,52.3305015563965,-0.2255620062351,234,10,20566,4068,0,1,'
             '2008-04-05 21:32:59,2008-04-05 21:32:59')
        expect(data[2]) == \
            ('23008,52.3506011962891,-0.2234109938145,234,10,10566,4068,0,1,'
             '2008-04-05 21:32:59,2008-04-05 21:32:59')

    def test_import_locations(self):
        expect(self.cells['22747']) == \
            Cell(22747, 52.0438995361328, -0.224637001752853, 234, 33, 2319,
                 647, 0, 1, datetime.datetime(2008, 4, 5, 21, 32, 40),
                 datetime.datetime(2008, 4, 5, 21, 32, 40))



#
# coding=utf-8
"""cellid - Imports OpenCellID data files"""
# Copyright © 2007-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import logging

from operator import attrgetter

from upoints import (point, utils)


class Cell(point.Point):

    """Class for representing a cellular cite from OpenCellID.org_

    .. versionadded:: 0.11.0

    .. _OpenCellID.org: http://opencellid.org/

    """

    __slots__ = ('ident', 'mcc', 'mnc', 'lac', 'cellid', 'crange', 'samples',
                 'created', 'updated')

    def __init__(self, ident, latitude, longitude, mcc, mnc, lac, cellid,
                 crange, samples, created, updated):
        """Initialise a new ``Cell`` object.

        :param int ident: OpenCellID database identifier
        :param float latitude: Cell's latitude
        :param float longitude: Cell's longitude
        :param int mcc: Cell's country code
        :param int mnc: Cell's network code
        :param int lac: Cell's local area code
        :param int cellid: Cell's identifier
        :param int crange: Cell's range
        :param int samples: Number of samples for the cell
        :param datetime.datetime created: Date the cell was first entered
        :param datetime.datetime updated: Date of the last update

        """
        super(Cell, self).__init__(latitude, longitude)
        self.ident = ident
        self.mcc = mcc
        self.mnc = mnc
        self.lac = lac
        self.cellid = cellid
        self.crange = crange
        self.samples = samples
        self.created = created
        self.updated = updated

    def __str__(self):
        """OpenCellID.org-style location string.

        .. seealso::

           :class:`point.Point`

        :rtype: ``str``
        :return: OpenCellID.org-style string representation of ``Cell`` object

        """
        return '%i,%.13f,%.13f,%i,%i,%i,%i,%i,%i,%s,%s' \
            % (self.ident, self.latitude, self.longitude, self.mcc, self.mnc,
               self.lac, self.cellid, self.crange, self.samples,
               self.created.strftime('%Y-%m-%d %H:%M:%S'),
               self.updated.strftime('%Y-%m-%d %H:%M:%S'))


class Cells(point.KeyedPoints):

    """Class for representing a group of :class:`Cell` objects.

    .. versionadded:: 0.11.0

    """

    def __init__(self, cells_file=None):
        """Initialise a new ``Cells`` object."""
        super(Cells, self).__init__()
        self._cells_file = cells_file
        if cells_file:
            self.import_locations(cells_file)

    def __str__(self):
        """``Cells`` objects rendered as export from OpenCellID.org.

        :rtype: ``str``
        :return: OpenCellID.org formatted output

        """
        return '\n'.join(map(str, sorted(self.values(),
                                         key=attrgetter('ident'))))

    def import_locations(self, cells_file):
        """Parse OpenCellID.org data files.

        ``import_locations()`` returns a dictionary with keys containing the
        OpenCellID.org_ database identifier, and values consisting of
        a ``Cell`` objects.

        It expects cell files in the following format::

            22747,52.0438995361328,-0.2246370017529,234,33,2319,647,0,1,
            2008-04-05 21:32:40,2008-04-05 21:32:40
            22995,52.3305015563965,-0.2255620062351,234,10,20566,4068,0,1,
            2008-04-05 21:32:59,2008-04-05 21:32:59
            23008,52.3506011962891,-0.2234109938145,234,10,10566,4068,0,1,
            2008-04-05 21:32:59,2008-04-05 21:32:59

        The above file processed by ``import_locations()`` will return the
        following ``dict`` object::

            {23008: Cell(23008, 52.3506011963, -0.223410993814, 234, 10, 10566,
                         4068, 0, 1, datetime.datetime(2008, 4, 5, 21, 32, 59),
                         datetime.datetime(2008, 4, 5, 21, 32, 59)),
             22747: Cell(22747, 52.0438995361, -0.224637001753, 234, 33, 2319,
                         647, 0, 1, datetime.datetime(2008, 4, 5, 21, 32, 40),
                         datetime.datetime(2008, 4, 5, 21, 32, 40)),
             22995: Cell(22995, 52.3305015564, -0.225562006235, 234, 10, 20566,
                         4068, 0, 1, datetime.datetime(2008, 4, 5, 21, 32, 59),
                         datetime.datetime(2008, 4, 5, 21, 32, 59))}

        :type cells_file: ``file``, ``list`` or ``str``
        :param cells_file: Cell data to read
        :rtype: ``dict``
        :return: Cell data with their associated database identifier

        .. _OpenCellID.org: http://opencellid.org/

        """
        self._cells_file = cells_file
        field_names = ('ident', 'latitude', 'longitude', 'mcc', 'mnc', 'lac',
                       'cellid', 'crange', 'samples', 'created', 'updated')
        parse_date = lambda s: datetime.datetime.strptime(s,
                                                          '%Y-%m-%d %H:%M:%S')
        field_parsers = (int, float, float, int, int, int, int, int, int,
                         parse_date, parse_date)
        data = utils.prepare_csv_read(cells_file, field_names)

        for row in data:
            try:
                cell = dict((n, p(row[n]))
                            for n, p in zip(field_names, field_parsers))
            except ValueError:
                if r"\N" in row.values():
                    # A few entries are incomplete, and when that occurs the
                    # export includes the string "\N" to denote missing
                    # data.  We just ignore them for now
                    logging.debug('Skipping incomplete entry %r' % row)
                    break
                else:
                    raise utils.FileFormatError('opencellid.org')
            else:
                self[row['ident']] = Cell(**cell)
